#include "RenderWidget.h"
#include <QDebug>
#include <QMouseEvent>


RenderWidget::RenderWidget(QWidget *parent) :
	QOpenGLWidget(parent)
	, m_error(false)
	//, m_models(filepaths.size())
{
	xRot = yRot = zRot = 0;

	QSurfaceFormat fmt = format();
	fmt.setSamples(16);
	setFormat(fmt);

	m_edgeCompression = true;
	m_directionalConstraint = true;
	m_globalAngle = 15.0f;
	m_globalAlpha = 0.0f;
	m_zNear = 0.3f;
	m_zFar = 100.0f;
}

void RenderWidget::initializeGL()
{
	qDebug() << "RenderWidget: initializeGL()";

	this->initializeOpenGLFunctions();

	setupShaders();
	setupLightingAndMatrices();
	setupBuffers();


	glEnable(GL_DEPTH_TEST); // needs to be enabled all the time
	glClearColor(.9f, .9f, .93f, 1.0f);

	initialized = true;
	qDebug() << "Render Widget: is initialized!";
}

void RenderWidget::setCutable(QString modelPath, bool cutable) {
	
	int pos = std::find(m_filepaths.begin(), m_filepaths.end(), modelPath) - m_filepaths.begin();
	qDebug() << "position of cut-able element: " << pos;
	if (pos < m_filepaths.size()) { // should always be the case, since we know what we rendered so it has to be in the vector
		m_models.at(pos)->setCutable(cutable);
	}
}

void RenderWidget::setCutAngle(int angle)
{
	m_globalAngle = angle;
}

void RenderWidget::setCutAreaAlpha(float alpha)
{
	m_globalAlpha = alpha;
}

void RenderWidget::setEdgeCompression(bool edgeComp)
{
	m_edgeCompression = edgeComp;
}

void RenderWidget::setDirectionalConstraint(bool dirConst)
{
	m_directionalConstraint = dirConst;
}

bool RenderWidget::loadModelsFromPaths() {
	qDebug() << "RenderWidget: loadModelsFromPaths()";
	if (m_filepaths.empty()) { // there are no models to render
		return false;
	}

	for (int i = 0; i < m_filepaths.size(); i++)
	{
		Model* m = new Model(m_colorShader, m_filepaths.at(i));
		m_models.at(i) = m;

	}

	modelsLoaded = true;
	return true;
}

bool RenderWidget::isInitialized() {
	return initialized;
}

void RenderWidget::setFilepaths(std::vector<QString> filepaths) {
	if (!this->m_filepaths.empty()) {
		this->m_filepaths.clear();
	}
	this->m_filepaths = filepaths;
	this->m_models.resize(m_filepaths.size());
	modelsLoaded = false;
}

void RenderWidget::mousePressEvent(QMouseEvent *event) {
	lastMousePos = event->pos();
}

// zooming with view matrix along z-axis
// and rotation according to mouse position
void RenderWidget::wheelEvent(QWheelEvent *event)
{
	QPoint degree = event->angleDelta(); // returns QPoint(0, 120) for forward and QPoint(0, -120 for backward rotation
	QPoint mousePos = event->pos(); // position for mouse position dependent zoom
	//qDebug() << mousePos;

	m_view.setToIdentity();

	if (degree.y() > 0) { // zoom in
		camPosZ = camPosZ - zoom;
		m_view.lookAt(QVector3D(0, 0,camPosZ),    
			QVector3D(0,0,0),
			QVector3D(0.0f, 1.0f, 0.0f));
		qDebug() << camPosZ << "zoom in";

		update();
	}
	else if (degree.y() < 0) { // zoom out
		camPosZ = camPosZ + zoom;
		m_view.lookAt(QVector3D(0, 0, camPosZ),
			QVector3D(0, 0, 0),
			QVector3D(0.0f, 1.0f, 0.0f));
		qDebug() << camPosZ << "zoom out";

		update();
	}
}

void RenderWidget::mouseMoveEvent(QMouseEvent *event)
{
	int dx = event->x() - lastMousePos.x();
	int dy = event->y() - lastMousePos.y();
	if (event->buttons() & Qt::LeftButton) {
		setXRotation(xRot + 8 * dy);
		setYRotation(yRot + 8 * dx);
	}
	else if (event->buttons() & Qt::RightButton) {
		setXRotation(xRot + 8 * dy);
		setZRotation(zRot + 8 * dx);
	}
	lastMousePos = event->pos();
}
static void qNormalizeAngle(int &angle) {
	while (angle < 0)
		angle = angle + 360 * 16;
	while (angle > 360)
		angle = angle - 360 * 16;
}
void RenderWidget::setXRotation(int angle) {
	qNormalizeAngle(angle);
	if (angle != xRot) {
		xRot = angle;
		update();
	}
}
void RenderWidget::setYRotation(int angle) {
	qNormalizeAngle(angle);
	if (angle != yRot) {
		yRot = angle;
		update();
	}
}
void RenderWidget::setZRotation(int angle) {
	qNormalizeAngle(angle);
	if (angle != zRot) {
		zRot = angle;
		update();
	}
}

QOpenGLShaderProgram* RenderWidget::createShaderProgram(QString vShader, QString fShader)
{
	QOpenGLShaderProgram* program = new QOpenGLShaderProgram();

	// Compile vertex shader
	if (!m_error && !program->addShaderFromSourceFile(QOpenGLShader::Vertex, vShader.toUtf8())) {
		qDebug() << "Error in vertex shader:" << program->log();
		m_error = true;
	}

	// Compile fragment shader
	if (!m_error && !program->addShaderFromSourceFile(QOpenGLShader::Fragment, fShader.toUtf8())) {
		qDebug() << "Error in fragment shader:" << program->log();
		m_error = true;
	}

	// Link the shader together into a program
	if (!m_error && !program->link()) {
		qDebug() << "Error linking shader program:" << program->log();
		m_error = true;
	}

	return program;
}


void RenderWidget::setupLightingAndMatrices()
{
	m_view.setToIdentity();
	m_view.lookAt(
		QVector3D(0.0f, 0.0f, camPosZ),    // Camera Position
		QVector3D(0.0f, 0.0f, 0.0f),    // Point camera looks towards
		QVector3D(0.0f, 1.0f, 0.0f));   // Up vector

	float aspect = 4.0f / 3.0f;
	m_projection.setToIdentity();
	m_projection.perspective(
		60.0f,          // field of vision
		aspect,         // aspect ratio
		m_zNear,		// near clipping plane
		m_zFar);		// far clipping plane

	m_lightInfo.Position = QVector4D(-1.0f, 1.0f, 1.0f, 1.0f);
	m_lightInfo.Intensity = QVector3D(1.0f, 1.0f, 1.0f);

}


void RenderWidget::setupShaders()
{
	m_colorShader = createShaderProgram("shaders/ColorShader.vert", "shaders/ColorShader.frag");
	m_cutShader = createShaderProgram("shaders/CutShader.vert", "shaders/CutShader.frag");
	m_rearHullsShader = createShaderProgram("shaders/RearHulls.vert", "shaders/RearHulls.frag");
	m_jumpFloodingShader = createShaderProgram("shaders/JumpFlooding.vert", "shaders/JumpFlooding.frag");
	m_bufferShader = createShaderProgram("shaders/FrameBuffer.vert", "shaders/FrameBuffer.frag");
	m_alphaBlendShader = createShaderProgram("shaders/AlphaBlend.vert", "shaders/AlphaBlend.frag");
}

void RenderWidget::setupBuffers()
{
	m_pingPongBuffers[0] = new CutawayBuffer(QWidget::width(), QWidget::height());
	m_pingPongBuffers[1] = new CutawayBuffer(QWidget::width(), QWidget::height());
	m_sceneBuffer = new CutawayBuffer(QWidget::width(), QWidget::height());
	m_cutAreasBuffer = new CutawayBuffer(QWidget::width(), QWidget::height());
}


void RenderWidget::resizeGL(int w, int h)
{
	glViewport(0, 0, w, h);

	m_projection.setToIdentity();
	m_projection.perspective(60.0f, (float)w / h, m_zNear, m_zFar);
	m_pingPongBuffers[0]->resize(w, h);
	m_pingPongBuffers[1]->resize(w, h);
	m_sceneBuffer->resize(w, h);
	m_cutAreasBuffer->resize(w, h);
}

void RenderWidget::paintGL()
{
	if (m_error || m_filepaths.empty())
		return;
	if (!modelsLoaded) {
		loadModelsFromPaths(); // for updating scene at runtime
	}


	// Render rear hulls of objects of interest into depth buffer
	renderRearHulls();

	// Create cutaway surface to cut secondary objects against (get index of stored buffer data)
	int bufferIdx = createCutawaySurface();

	// Render secondary objects against cutaway surface into buffer
	renderSecondaryObjects(bufferIdx);

	// Render cut areas faded out with alpha blending
	renderCutAreas();


	// Render both buffers to screen via blending
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Enable blending
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


	m_sceneBuffer->bindColorTexToShader(m_bufferShader);
	m_sceneBuffer->draw();

	m_cutAreasBuffer->bindColorTexToShader(m_bufferShader);
	m_cutAreasBuffer->draw();
	
	m_bufferShader->release();
	
	// Disable blending
	glBlendFunc(GL_ONE, GL_ZERO);
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);


	// Render objects of interest above all 
	renderObjectsOfInterest();
}


void RenderWidget::renderSecondaryObjects(int bufferIdx)
{

	// Set shader uniforms
	m_cutShader->bind();
	m_cutShader->setUniformValue("imgSize", QWidget::width(), QWidget::height());
	m_cutShader->setUniformValue("zFar", m_zFar);
	m_cutShader->setUniformValue("zNear", m_zNear);
	m_cutShader->setUniformValue("lightPosition", m_lightInfo.Position);
	m_cutShader->setUniformValue("lightIntensity", m_lightInfo.Intensity);
	m_cutShader->setUniformValue("projection", m_projection);

	// Bind cutaway surface textures to shader
	m_pingPongBuffers[bufferIdx]->bindTexturesToShader(m_cutShader);

	//m_pingPongBuffers[bufferIdx]->bindTexturesToShader(m_bufferShader);
	//m_pingPongBuffers[bufferIdx]->draw();
	//m_bufferShader->release();
	
	m_sceneBuffer->bind();

	glClearColor(.9f, .9f, .93f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Draw secondary objects against cutaway surface
	for (int i = 0; i < m_models.size(); i++)
	{
		if (m_models.at(i)->isCutable())
		{
			m_models.at(i)->draw(m_cutShader, getTransform(), m_view, m_projection);
		}
	}

	m_sceneBuffer->unbind();
	m_cutShader->release();

}

void RenderWidget::renderCutAreas()
{

	// Set shader uniforms
	m_alphaBlendShader->bind();
	m_alphaBlendShader->setUniformValue("imgSize", QWidget::width(), QWidget::height());
	m_alphaBlendShader->setUniformValue("zFar", m_zFar);
	m_alphaBlendShader->setUniformValue("zNear", m_zNear);
	m_alphaBlendShader->setUniformValue("lightPosition", m_lightInfo.Position);
	m_alphaBlendShader->setUniformValue("lightIntensity", m_lightInfo.Intensity);
	m_alphaBlendShader->setUniformValue("maxAlpha", m_globalAlpha);

	// Bind scene and cutaway surface depth map to shader
	m_sceneBuffer->bindTexturesToShader(m_alphaBlendShader);

	m_cutAreasBuffer->bind();

	// Clear color and depth
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// Draw secondary (cut-able) objects
	for (int i = 0; i < m_models.size(); i++)
	{
		if (m_models.at(i)->isCutable())
		{
			m_models.at(i)->draw(m_alphaBlendShader, getTransform(), m_view, m_projection);
		}
	}

	m_cutAreasBuffer->unbind();
	m_alphaBlendShader->release();

}

void RenderWidget::renderObjectsOfInterest()
{
	// Set shader uniforms
	m_colorShader->bind();
	m_colorShader->setUniformValue("imgSize", QWidget::width(), QWidget::height());
	m_colorShader->setUniformValue("lightPosition", m_lightInfo.Position);
	m_colorShader->setUniformValue("lightIntensity", m_lightInfo.Intensity);


	// Draw objects of interest
	for (int i = 0; i < m_models.size(); i++)
	{
		if (!m_models.at(i)->isCutable())
		{
			m_models.at(i)->draw(m_colorShader, getTransform(), m_view, m_projection);
		}
	}

	m_colorShader->release();
}


void RenderWidget::renderRearHulls()
{

	// Settings to get rear hulls of objects of interest
	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClearDepth(0.0);			// initialize buffers with 0
	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);		// cull front faces away
	glDepthFunc(GL_GREATER);	// render meshes furthest away

	// Clear both buffer and render objects of interest to buffer 0
	m_pingPongBuffers[1]->bind();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	m_pingPongBuffers[1]->unbind();

	m_pingPongBuffers[0]->bind();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Bind shader program
	m_rearHullsShader->bind();
	m_rearHullsShader->setUniformValue("invView", m_view.inverted());
	m_rearHullsShader->setUniformValue("imgSize", QWidget::width(), QWidget::height());
	m_rearHullsShader->setUniformValue("theta", (float)(m_globalAngle * M_PI / 180.0f)); // radians = degree * PI / 180
	m_rearHullsShader->setUniformValue("zNear", m_zNear);
	m_rearHullsShader->setUniformValue("zFar", m_zFar);
	m_rearHullsShader->setUniformValue("edgeCompression", m_edgeCompression);
	m_rearHullsShader->setUniformValue("directionalConstraint", m_directionalConstraint);


	// Draw all objects of interest
	for (int i = 0; i < m_models.size(); i++)
	{
		if (!m_models.at(i)->isCutable())
		{
			m_models.at(i)->draw(m_rearHullsShader, getTransform(), m_view, m_projection);
		}
		
	}

	m_pingPongBuffers[0]->unbind();
	m_rearHullsShader->release();

	// Reset OpenGL settings
	glClearDepth(1.0);
	glDepthFunc(GL_LESS);
	glDisable(GL_CULL_FACE);

}


int RenderWidget::createCutawaySurface()
{
	glDepthMask(GL_TRUE);	// depth buffer is enabled for writing (default setting)
	glDepthFunc(GL_ALWAYS); // specify the value used for depth buffer comparisons - passes always!


	// Number of iterations of nxn Image: N = log2(n)
	QVector2D imageSize(QWidget::width(), QWidget::height());
	int n = std::max(imageSize.x(), imageSize.y()); // our image is not symmetric so we use its max length for n
	int N = (int)ceil( log2(n) );

	// Jump length k changes every iteration (k = n/2^i) 
	int k;

	// Alternating input and output buffer
	int inIdx, outIdx;

	// Set uniforms
	m_jumpFloodingShader->bind();
	m_jumpFloodingShader->setUniformValue("imgSize", imageSize.x(), imageSize.y() );
	m_jumpFloodingShader->setUniformValue("theta", (float)(m_globalAngle * M_PI / 180.0f)); // radians = degree * PI / 180
	m_jumpFloodingShader->setUniformValue("zNear", m_zNear);
	m_jumpFloodingShader->setUniformValue("zFar", m_zFar);
	m_jumpFloodingShader->setUniformValue("final", false);


	// Apply jump flooding algorithm on rear hulls texture
	for (int i = 0; i < N; i++)
	{
		// Switch buffer every iteration
		inIdx = i % 2; // 0,1,0,...
		outIdx = (i + 1) % 2; // 1,0,1,...

		m_pingPongBuffers[outIdx]->bind();

		// Set uniforms and draw buffer quad
		if (i == N - 1)
			m_jumpFloodingShader->setUniformValue("final", true);

		k = std::ceil(n / pow(2, i + 1)); // k = 1/2, 1/4, 1/8, ... 1/2^N
		m_jumpFloodingShader->setUniformValue("k", k);

		m_pingPongBuffers[inIdx]->bindTexturesToShader(m_jumpFloodingShader);
		m_pingPongBuffers[inIdx]->draw();


		m_pingPongBuffers[outIdx]->unbind();

	}
	
	m_jumpFloodingShader->release();


	glDepthFunc(GL_LESS);
	
	return outIdx;
}

QMatrix4x4 RenderWidget::getTransform()
{
	// Set the transform matrix
	QMatrix4x4 transform;
	transform.setToIdentity();
	transform.translate(-0.2f, -0.5f, .5f);
	transform.scale(0.25f);
	transform.rotate(xRot / 16.0f, 1.0f, 0.0f, 0.0f);
	transform.rotate(yRot / 16.0f, 0.0f, 1.0f, 0.0f);
	transform.rotate(zRot / 16.0f, 0.0f, 0.0f, 1.0f);

	return transform;
}




void RenderWidget::closeEvent(QCloseEvent *event) {
	qDebug() << "closeEvent()";
	QOpenGLWidget::closeEvent(event);
}

RenderWidget::~RenderWidget() {
	
	m_colorShader->release();
	m_colorShader->removeAllShaders();	
	
	m_cutShader->release();
	m_cutShader->removeAllShaders();

	m_rearHullsShader->release();
	m_rearHullsShader->removeAllShaders();

	m_jumpFloodingShader->release();
	m_jumpFloodingShader->removeAllShaders();

	m_bufferShader->release();
	m_bufferShader->removeAllShaders();

	m_alphaBlendShader->release();
	m_alphaBlendShader->removeAllShaders();


	delete m_pingPongBuffers[0]; m_pingPongBuffers[0] = nullptr;
	delete m_pingPongBuffers[1]; m_pingPongBuffers[1] = nullptr;
	delete m_sceneBuffer; m_sceneBuffer = nullptr;
	delete m_cutAreasBuffer; m_cutAreasBuffer = nullptr;
}