#pragma once

#define _USE_MATH_DEFINES // for C++  
#include <cmath> 

#include <vector>
#include <algorithm>
#include <iterator>
#include <string>
#include <QOpenGLShaderProgram>
#include <QOpenGLBuffer>
#include <QOpenGLVertexArrayObject>
#include <QOpenGLFunctions>
#include <QOpenGLWidget>
#include <QPoint>

#include "loader/Model.h"
#include "cutaway/CutawayBuffer.h"

/**
* RenderWidget represents the area where the scene is dynamically rendered to.
* It executes the main render processes including the cutaway functionality.
*/
class RenderWidget: public QOpenGLFunctions, public QOpenGLWidget
{
	
public:
	RenderWidget(QWidget *parent);
	~RenderWidget();

	/**
	* Sets file paths of all models of the current scene to be able to load them.
	* @see loadModelsFromPaths()
	*/
	void setFilepaths(std::vector<QString> filepaths);

	/**
	* Load models of the current scene with file paths.
	* @see setFilepaths()
	*/
	bool loadModelsFromPaths();

	/**
	* Returns true if the RenderWidget is initialized. False otherwise.
	*/
	bool isInitialized();

	/**
	* Sets the model (un-)cut-able.
	* @param modelPath the path of the model which is set (un-)cut-able
	* @param cut-able enables/disables the model to be cut
	*/
	void setCutable(QString modelPath, bool cutable);

	/**
	* Sets the global cutaway angle.
	* @param angle specifies the angle of cutting
	*/
	void setCutAngle(int angle);

	/**
	* Sets the global alpha value for the areas cut away.
	* @param alpha the used alpha value
	*/
	void setCutAreaAlpha(float alpha);

	/**
	* Enables or disables edge compression.
	*/
	void setEdgeCompression(bool edgeComp);

	/**
	* Enables or disables edge compression.
	*/
	void setDirectionalConstraint(bool dirConst);

protected:

	/**
	* Implementation of the parent function of QOpenGLWidget. 
	* It initializes OpenGL functions and executes the setup of all necessary render components.
	*/
	void initializeGL();

	/**
	* Implementation of the parent function of QOpenGLWidget.
	* It resizes the widget and all frame buffers appropriately. OpenGL functions and executes the setup of all necessary render components.
	*/
	void resizeGL(int w, int h);

	/**
	* Implementation of the parent function of QOpenGLWidget.
	* It executes the per frame rendering by calling all render methods and does the frame buffer blending between the cut and uncut scene.
	* @see renderRearHulls()
	* @see renderSecondaryObjects()
	* @see renderCuttedAreas()
	* @see renderObjectsOfInterest()
	*/
	void paintGL();

	/**
	* Implementation of the parent function of QOpenGLWidget.
	* Calls the close event function of QOpenGLWidget.
	*/
	void closeEvent(QCloseEvent *event) override;

	/**
	* Processes the mouse movement of the user by manipulating the global rotation angles of x, y and z directions.
	*/
	void mouseMoveEvent(QMouseEvent *event) override;

	/**
	* Responsible for zooming in and out. 
	* Processes the mouse wheel interaction of the user by manipulating the camera position (view matrix) appropriately.
	*/
	void wheelEvent(QWheelEvent *event) override;

	/**
	* Saves the current mouse position for mouse movement processing.
	*/
	void mousePressEvent(QMouseEvent *event) override;


private:

	// Initialize
	/**
	* Compiles and links the two shader together into a shader program.
	* @param vShader path to the vertex shader
	* @param fShader path to the fragment shader
	* @return the linked shader program
	*/
	QOpenGLShaderProgram* createShaderProgram(QString vShader, QString fShader);

	/**
	* Initializes the view, projection and light matrices needed for rendering.
	*/
	void setupLightingAndMatrices();

	/**
	* Creates all necessary shader programs to render the scene.
	* @see createShaderProgram()
	*/
	void setupShaders();

	/**
	* Creates all frame buffers needed for rendering. 
	* Two "ping-pong" buffers for the jump flooding algorithm and one for the secondary objects of the scene and one for the cut areas.
	*/
	void setupBuffers();


	// Rendering
	/**
	* Renders the rear (back) hulls of the objects of interest into the first "ping-pong" buffer.
	* This is the first step for creating the cutaway surface.
	*/
	void renderRearHulls();

	/**
	* Renders all secondary objects (cut-able) of the scene against the cutaway surface and into a frame buffer.
	* @param bufferIdx index of the "ping-pong" buffer containing the final cutaway surface.
	*/
	void renderSecondaryObjects(int bufferIdx);

	/**
	* Renders the cut areas clipped according the cutaway surfaces into a frame buffer.
	* It sets the cut areas transparency according the global alpha value.
	*/
	void renderCutAreas();

	/**
	* Renders all objects of interest (uncut-able) above all directly to the screen.
	*/
	void renderObjectsOfInterest();

	/**
	* Creates the cutaway surface the scene is clipped against.
	* It uses the rear hulls of the objects of interest an executes the jump flooding algorithm to create the final cut volume.
	* @return index of the "ping-pong" buffer containing the final cutaway surface
	* @see renderRearHulls()
	*/
	int createCutawaySurface();

	/**
	* Calculates the global transformation matrix by using the mouse rotation made by the user.
	* @return transformation matrix
	*/
	QMatrix4x4 getTransform();

	/**
	* Sets the global rotation in x direction.
	* @param angle rotation angle
	*/
	void setXRotation(int angle);

	/**
	* Sets the global rotation in y direction.
	* @param angle rotation angle
	*/
	void setYRotation(int angle);

	/**
	* Sets the global rotation in z direction.
	* @param angle rotation angle
	*/
	void setZRotation(int angle);


	float camPosZ = 3.2f;
	float zoom = 0.5f;

	// Model
	std::vector<Model*> m_models;
	QMatrix4x4 m_projection, m_view;
	std::vector<QString> m_filepaths;
	LightInfo m_lightInfo;

	// Buffer
	CutawayBuffer* m_pingPongBuffers[2];
	CutawayBuffer* m_sceneBuffer;
	CutawayBuffer* m_cutAreasBuffer;

	// Shader
	QOpenGLShaderProgram* m_colorShader;
	QOpenGLShaderProgram* m_cutShader;
	QOpenGLShaderProgram* m_rearHullsShader;
	QOpenGLShaderProgram* m_jumpFloodingShader;
	QOpenGLShaderProgram* m_bufferShader;
	QOpenGLShaderProgram* m_alphaBlendShader;

	float m_globalAngle; // cutaway angle in degree
	float m_globalAlpha; // alpha value used for cut areas

	float m_zNear, m_zFar;

	QPoint lastMousePos;
	int xRot, yRot, zRot;

	bool initialized = false;
	bool modelsLoaded;
	bool m_error;
	bool m_edgeCompression;
	bool m_directionalConstraint;

};