//reference: http://learnopengl.com/#!Advanced-OpenGL/Framebuffers

#pragma once

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <iostream>

#include <QOpenGLFunctions>
#include <QOpenGLShaderProgram>
#include <QOpenGLVertexArrayObject>
#include <QOpenGLFramebufferObject>
#include <QOpenGLBuffer>

#define S_QUAD_VERTEX_COUNT 24


/**
* Represents a custom frame buffer. 
* It is used to render to the textures of this buffer and to do the cutaway calculations on GPU side by writing from and to its textures.
*/
class CutawayBuffer : public QOpenGLFunctions
{
public:

	/**
	* Constructor for CutawayBuffer. 
	* @param screenWidth the width used to calculate the frame buffer width
	* @param screenHeight the height used to calculate the frame buffer height
	*/
	CutawayBuffer(int screenWidth, int sreenHeight);
	~CutawayBuffer();

	/**
	* Binds the frame buffer to be able to render to its textures instead to the default OpenGL frame buffer.
	*/
	void bind();

	/**
	* Binds the color and depth attachment of the frame buffer to the given shader.
	* @param shader the shader program to which the color and depth textures are bound
	* @param textureUnit the unit at which the texture is bound to the shader (default is 0)
	*/
	void bindTexturesToShader(QOpenGLShaderProgram* shader, int textureUnit = 0);

	/**
	* Binds only the color attachment of the frame buffer to the given shader.
	* @param shader the shader program to which the color texture is bound
	*/
	void bindColorTexToShader(QOpenGLShaderProgram* shader);

	/**
	* Draws the frame buffer to the screen.
	* If this is called the first time a vertex buffer which spans the screen quad is created. 
	* This buffer is bound to a vertex object array to enable fast rendering.
	*/
	void draw();

	/**
	* Unbinds the frame buffer to be able to render to the default OpenGL frame buffer again.
	*/
	void unbind();

	/**
	* Is called if a new size of the frame buffer is needed (e.g. if the window resizes).
	* It calls init() to create new color and depth attachments by using the new size.
	* @param width new frame buffer width
	* @parma height new frame buffer height
	* @see init()
	*/
	void resize(int width, int height);

private:

	int m_width, m_height;

	QOpenGLShaderProgram * m_shader;
	QOpenGLVertexArrayObject m_vao;	// Vertex array buffer object
	QOpenGLBuffer m_vbo;			// Vertex buffer object

	GLuint m_colorTexture;
	GLuint m_depthTexture;
	GLuint m_fbo;

	/**
	* Initializes the frame buffer with a color and a depth attachment.
	*/
	void init();


};