#include "Model.h"



Model::Model(QOpenGLShaderProgram* shaderProgram, QString filepath) :
	m_filepath(filepath)
	, m_shaderProgram(shaderProgram)
	, m_indexBuffer(QOpenGLBuffer::IndexBuffer)
	, m_model(QMatrix4x4())
	, m_cutable(true)

{
	qDebug() << "Initialize MODEL";

	this->initializeOpenGLFunctions();

	createBuffers();
	createAttributes();

	m_materialInfo.Ambient = QVector3D(0.3f, 0.3f, 0.3f);
	m_materialInfo.Diffuse = QVector3D(0.4f, 0.4f, 0.4f);
	m_materialInfo.Specular = QVector3D(0.5f, 0.5f, 0.5f);
	m_materialInfo.Shininess = 50.0f;
}


Model::~Model()
{
	if (m_vertexBuffer.isCreated())
		m_vertexBuffer.destroy();

	if (m_normalBuffer.isCreated())
		m_normalBuffer.destroy();

	if (m_textureUVBuffer.isCreated())
		m_textureUVBuffer.destroy();

	if (m_indexBuffer.isCreated())
		m_indexBuffer.destroy();

	if (m_vao.isCreated())
		m_vao.destroy();
}


void Model::createBuffers()
{
	qDebug() << "Create Buffers ";

	ModelLoader model;
	if (!model.Load(m_filepath))
	{
		qDebug() << "Could not load model!!!";
		return;
	}

	QVector<float> *vertices;
	QVector<float> *normals;
	QVector<QVector<float> > *textureUV;
	QVector<unsigned int> *indices;

	model.getBufferData(&vertices, &normals, &indices);
	model.getTextureData(&textureUV, 0, 0);

	// Create a vertex array object
	m_vao.create();
	m_vao.bind();

	// Create a buffer and copy the vertex data to it
	m_vertexBuffer.create();
	m_vertexBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
	m_vertexBuffer.bind();
	m_vertexBuffer.allocate(&(*vertices)[0], vertices->size() * sizeof(float));

	// Create a buffer and copy the vertex data to it
	m_normalBuffer.create();
	m_normalBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
	m_normalBuffer.bind();
	m_normalBuffer.allocate(&(*normals)[0], normals->size() * sizeof(float));

	if (textureUV != 0 && textureUV->size() != 0)
	{
		// Create a buffer and copy the vertex data to it
		m_textureUVBuffer.create();
		m_textureUVBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
		m_textureUVBuffer.bind();
		int texSize = 0;
		for (int ii = 0; ii<textureUV->size(); ++ii)
			texSize += textureUV->at(ii).size();

		m_textureUVBuffer.allocate(&(*textureUV)[0][0], texSize * sizeof(float));
	}

	// Create a buffer and copy the index data to it
	m_indexBuffer.create();
	m_indexBuffer.setUsagePattern(QOpenGLBuffer::StaticDraw);
	m_indexBuffer.bind();
	m_indexBuffer.allocate(&(*indices)[0], indices->size() * sizeof(unsigned int));

	m_rootNode = model.getNodeData();
	
}



void Model::createAttributes()
{

	m_vao.bind();

	// Set up the vertex array state
	m_shaderProgram->bind();


	// Map vertex data to the vertex shader's layout location '0'
	m_vertexBuffer.bind();
	m_shaderProgram->enableAttributeArray(0);      // layout location
	m_shaderProgram->setAttributeBuffer(0,          // layout location
		GL_FLOAT,   // data's type
		0,          // Offset to data in buffer
		3);         // number of components (3 for x,y,z)

	// Map normal data to the vertex shader's layout location '1'
	m_normalBuffer.bind();
	m_shaderProgram->enableAttributeArray(1);      // layout location
	m_shaderProgram->setAttributeBuffer(1,          // layout location
		GL_FLOAT,   // data's type
		0,          // Offset to data in buffer
		3);         // number of components (3 for x,y,z)


	if (m_textureUVBuffer.isCreated()) {
		// Map texture coordinate data to the vertex shader's layout location '2'
		m_textureUVBuffer.bind();
		m_shaderProgram->enableAttributeArray(2);      // layout location
		m_shaderProgram->setAttributeBuffer(2,          // layout location
			GL_FLOAT,   // data's type
			0,          // Offset to data in buffer
			2);         // number of components (2 for u,v)
	}

}



void Model::draw(QOpenGLShaderProgram* shader, QMatrix4x4 transform, QMatrix4x4 view, QMatrix4x4 projection)
{
	if (m_shaderProgram != shader)
	{
		m_shaderProgram = shader;
		createAttributes();
	}

	m_view = view;
	m_projection = projection;
	m_transform = transform;

	// Bind VAO and draw everything
	m_vao.bind();
	drawNode(m_rootNode.data(), QMatrix4x4());
	m_vao.release();

	// Unbind texture
	glBindTexture(GL_TEXTURE_2D, 0);

}



void Model::drawNode(const Node *node, QMatrix4x4 objectMatrix)
{
	// Prepare matrices
	objectMatrix *= node->transformation;
	QMatrix4x4 modelMatrix = m_model * m_transform * objectMatrix;
	QMatrix4x4 modelViewMatrix = m_view * modelMatrix;
	QMatrix3x3 normalMatrix = modelViewMatrix.normalMatrix();
	QMatrix4x4 mvp = m_projection * modelViewMatrix;

	m_shaderProgram->setUniformValue("MV", modelViewMatrix);// Transforming to eye space
	m_shaderProgram->setUniformValue("N", normalMatrix);    // Transform normal to Eye space
	m_shaderProgram->setUniformValue("MVP", mvp);           // Matrix for transforming to Clip space

	// Draw each mesh in this node
	for (int imm = 0; imm<node->meshes.size(); ++imm)
	{
		if (node->meshes[imm]->material->Name == QString("DefaultMaterial"))
			setMaterialUniforms(m_materialInfo);
		else
			setMaterialUniforms(*node->meshes[imm]->material);

		glDrawElements(GL_TRIANGLES, node->meshes[imm]->indexCount, GL_UNSIGNED_INT
			, (const void*)(node->meshes[imm]->indexOffset * sizeof(unsigned int)));
	}

	// Recursively draw this nodes children nodes
	for (int inn = 0; inn<node->nodes.size(); ++inn)
		drawNode(&node->nodes[inn], objectMatrix);
}


void Model::setMaterialUniforms(MaterialInfo &mater)
{
	// Set texture and default material attributes
	if (mater.isTexture)
	{
		m_shaderProgram->setUniformValue("hasTexture", true);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, mater.textureId);
		m_shaderProgram->setUniformValue("Texture", 0); // GL_TEXTURE0
	}
	// Set custom material attributes
	else
	{
		m_shaderProgram->setUniformValue("hasTexture", false);
	}

	m_shaderProgram->setUniformValue("Ka", mater.Ambient);
	m_shaderProgram->setUniformValue("Kd", mater.Diffuse);
	m_shaderProgram->setUniformValue("Ks", mater.Specular);
	m_shaderProgram->setUniformValue("shininess", mater.Shininess);
}

void Model::scaleModel(float factor)
{
	m_model.scale(factor);
}

void Model::setCutable(bool cutable)
{
	m_cutable = cutable;
}

bool Model::isCutable()
{
	return m_cutable;
}