#pragma once

#include <QOpenGLBuffer>
#include <QOpenGLVertexArrayObject>
#include <QOpenGLShaderProgram>
#include <QString>
#include <QOpenGLFunctions>

#include "loader/modelloader.h"

/**
* This class represents an scene object. 
* It can be an object of interest or a secondary object depending on its cut-able state which can change dynamically.
*/
class Model : public QOpenGLFunctions
{
public:

	/**
	* Constructor of the Model class.
	* @param shaderProgram the default shader program to render this model
	* @param filepath the file path where the model can be loaded from
	*/
	Model(QOpenGLShaderProgram* shaderProgram, QString filepath);
	~Model();

	/**
	* Renders the model with the given shader and matrices by recursively calling drawNode().
	* @param shader shader program to render with
	* @param rotation matrix containing the rotation to apply before rendering
	* @param view matrix containing the camera transformation
	* @param projection matrix containing the projection
	* @see drawNode()
	*/
	void draw(QOpenGLShaderProgram* shader, QMatrix4x4 rotation, QMatrix4x4 view, QMatrix4x4 projection);

	/**
	* Returns the cut-able state of the model. If it is false it is an object of interest and a secondary object otherwise.
	* @return cut-able state
	*/
	bool isCutable();

	/**
	* Sets the cut-able state of this model.
	* @param cut-able the state to set
	*/
	void setCutable(bool cutable);

	/**
	* Scales the whole model with the same factor in all directions.
	* @param factor used for scaling
	*/
	void scaleModel(float factor);

private:

	bool m_cutable;

	QOpenGLShaderProgram* m_shaderProgram;

	QString m_filepath;

	QOpenGLVertexArrayObject m_vao;

	QOpenGLBuffer m_vertexBuffer;
	QOpenGLBuffer m_normalBuffer;
	QOpenGLBuffer m_textureUVBuffer;
	QOpenGLBuffer m_indexBuffer;

	QSharedPointer<Node> m_rootNode;

	QMatrix4x4  m_model, m_projection, m_view, m_transform;


	MaterialInfo m_materialInfo;

	/**
	* Creates and initializes buffers for vertices, normals, texture coordinates and indices.
	* These buffers are then bound to a vertex array object to speed up rendering.
	*/
	void createBuffers();

	/**
	* Create attribute arrays for all buffers to be able to access their data in shader.
	*/
	void createAttributes();

	/**
	* Renders recursively all nodes of this model starting at the root node.
	* @param node the current node to render
	* @param objectMatrix the local transformation matrix of this node
	*/
	void drawNode(const Node *node, QMatrix4x4 objectMatrix);

	/**
	* Sets the material attributes for the current shader. 
	* Uses the material information of the current node or the default settings if there is no such information.
	*/
	void setMaterialUniforms(MaterialInfo &mater);

};