#ifndef MODELLOADER_H
#define MODELLOADER_H

#include <string>
#include <QMatrix4x4>
#include <vector>
#include <QFile>
#include <QSharedPointer>
#include <QDir>
#include <QOpenGLFunctions>
#include "freeimage\x64\FreeImage.h"


struct aiScene;
struct aiNode;
struct aiMesh;
struct aiMaterial;


/**
* Struct containing all necessary material information.
*/
struct MaterialInfo
{
    QString Name;
    QVector3D Ambient;
    QVector3D Diffuse;
    QVector3D Specular;
    float Shininess;

    bool isTexture;
    QString textureName;
	GLint textureId;
};

/**
* Struct containing all necessary light information.
*/
struct LightInfo
{
    QVector4D Position;
    QVector3D Intensity;
};

/**
* Struct containing all necessary mesh information.
*/
struct Mesh
{
    QString name;
    unsigned int indexCount;
    unsigned int indexOffset;
    QSharedPointer<MaterialInfo> material;

    unsigned int numUVChannels;
    bool hasTangentsAndBitangents;
    bool hasNormals;
    bool hasBones;
};

/**
* Struct representing an Assimp node.
*/
struct Node
{
    QString name;

    QMatrix4x4 transformation;
    QVector<QSharedPointer<Mesh> > meshes;
    QVector<Node> nodes;
};

/**
* This class is responsible for model loading via Assimp and FreeImage libraries.
*/
class ModelLoader : public QOpenGLFunctions
{
public:
	/**
	* Constructor of the ModelLoader class.
	* @param transformToUnitCoordinates defines if the loaded model is transformed to unit coordinates or not
	*/
    ModelLoader(bool transformToUnitCoordinates = false);
   
	/**
	* Loads model at given directory.
	* @param filePath the file path to the model
	*/
	bool Load(QString filePath);

	/**
	* Writes vertex, normal and index data of the loaded model into the given vectors.
	* @param vertices empty vector to write the vertex data to
	* @param normals empty vector to write the normal data to
	* @param indices empty vector to write the index data to
	*/
    void getBufferData( QVector<float> **vertices, QVector<float> **normals, QVector<unsigned int> **indices);
	
	/**
	* Writes texture coordinates, tangents and bitangents of current model into the given vectors to enable texture and normal mapping.
	* @param texturUV empty vector to write the texture coordinates to
	* @param tangents empty vector to write the tangents to
	* @param bitangents empty vector to write the bitangents to
	*/
    void getTextureData( QVector<QVector<float> > **textureUV,                   // For texture mapping
                         QVector<float> **tangents, QVector<float> **bitangents);// For normal mapping

	/**
	* Returns the root node of the current model exported via Assimp library. 
	* @return a shared pointer to the root node
	*/
    QSharedPointer<Node> getNodeData() { return m_rootNode; }

    // Texture information
	/**
	* Returns the number of UV channels based on the texture of the material assigned to the model.
	* @return number of UV channels
	*/
    int numUVChannels() { return m_textureUV.size(); }

	/**
	* Returns the number of UV components based on the texture of the material assigned to the model.
	* @return number of UV components
	*/
    int numUVComponents(int channel) { return m_numUVComponents.at(channel); }

private:

	/**
	* Extracts the material information with ambient, diffuse and specular color and material shininess and stores it in the MaterialInfo struct.
	* @param mater Assimp material
	* @return material information stored in MaterialInfo struct
	*/
    QSharedPointer<MaterialInfo> processMaterial(aiMaterial *mater);

	/**
	* Extracts mesh information with vertices, normals, indices and texture information.
	* @param mesh the mesh to process
	* @return mesh information stored in Mesh struct
	*/
    QSharedPointer<Mesh> processMesh(aiMesh *mesh);

	/**
	* Recursive called method to process all Assimp nodes starting at the root node.
	* @param scene current Assimp scene
	* @param node Assimp node to get the data from
	* @param parentNode parent of the new node which will be created
	* @param newNode empty Node struct which will filled with information of the given Assimp node
	*/
    void processNode(const aiScene *scene, aiNode *node, Node *parentNode, Node &newNode);

	/**
	* Loads textures of the current model via FreeImage library.
	* @param path the path to the texture folder
	* @param directory the name of the texture
	* @return the texture ID to be able to bind this texture with OpenGL when rendering this model
	*/
	GLint TextureFromFile(const char* path, std::string directory);

	/**
	* Transform the model to unit coordinates, so a model of any size or shape will fit on screen.
	*/
    void transformToUnitCoordinates();

	/**
	* Used by transformToUnitCoordinates() to get the minimum and maximum x,y,z values for the model.
	* Is called recursive for each node.
	* @param node current node to process x,y,z values
	* @param transformation local transformation of the current node
	* @param minDimension empty vector to store the minimum values
	* @param maxDimension empty vector to store the maximum values
	* @see transformToUnitCoordinates()
	*/
    void findObjectDimensions(Node *node, QMatrix4x4 transformation, QVector3D &minDimension, QVector3D &maxDimension);

    QVector<float> m_vertices;
    QVector<float> m_normals;
    QVector<unsigned int> m_indices;

    QVector<QVector<float/*texture mapping coords*/> > m_textureUV; // m_textureUV[uvChannelIndex] is vector of texture mapping coords
                                                                    // m_textureUV[uvChannelIndex][ii+n] == if(n==0&&numCmpnts>0) U; if(n==0&&numCmpnts>0) V; if(n==0&&numCmpnts>0) W.
    QVector<float> m_tangents;
    QVector<float> m_bitangents;
    QVector<unsigned int /*num components*/> m_numUVComponents; // m_numUVComponents[uvChannelIndex]

    QVector<QSharedPointer<MaterialInfo> > m_materials;
    QVector<QSharedPointer<Mesh> > m_meshes;
    QSharedPointer<Node> m_rootNode;
    bool m_transformToUnitCoordinates;

	std::string m_directory;
};

#endif // MODELLOADER_H
