#version 430 core

uniform float zNear;
uniform float zFar;
uniform sampler2D depthMap0;
uniform vec2 imgSize;

// Light information
uniform vec4 lightPosition;
uniform vec3 lightIntensity;

// Material information
uniform bool hasTexture;
uniform sampler2D Texture;
uniform vec3 Ka;
uniform vec3 Kd;
uniform vec3 Ks;
uniform float shininess;

uniform mat4 projection;

// Input variables coming from vertex shader, interpolated to this fragment
in vec3 interpolatedPosition;
in vec3 interpolatedNormal;
in vec2 TexCoords;

out vec4 fragmentColor;

void main()
{
	
	// Check if fragment is inside the cutting volume -> discard
	float depth = -(interpolatedPosition.z + zNear) / (zFar - zNear); // -interpolatedPosition.z / zFar; // https://learnopengl.com/Advanced-OpenGL/Depth-testing
	
	vec2 texCoord = gl_FragCoord.xy / imgSize;
	float depthMapSample = texture(depthMap0, texCoord).r;
	
	if (depth < depthMapSample)
	{
		discard;
	}
	
	
	vec3 ambient  = Ka;
	vec3 diffuse = Kd;
	vec3 specular = Ks;
	
	if(!gl_FrontFacing) {
	
	// normal has been interpolated, so we need to normalize it
	float viewDepth = -(depthMapSample/(zFar-zNear) + zNear);//-depthMapSample * zFar;
	vec3 viewPosition = vec3(texCoord * (-viewDepth) / vec2(projection[0][0], projection[1][1]), viewDepth);

	
	vec3 normalVector = normalize(cross(dFdx(viewPosition), dFdy(viewPosition)));//normalize(-interpolatedNormal);

	// Calculate light source vector
	vec3 lightSourceVector = normalize( lightPosition.xyz - interpolatedPosition);

	// Calculate the view vector
	vec3 viewVector = normalize( -interpolatedPosition.xyz );

	// Ambient contribution
	vec3 ambientContribution = ambient;

	// Default black diffuse and specular contributions
	vec3 diffuseContribution = vec3(0.0);
	vec3 specularContribution = vec3(0.0);

	// Dot product of two normalized vectors gives us a value
	// indicating how close they are to each other,
	// A value of -1 means they are pointing in opposite directions
	// while a value of 1 means they are pointing in the same direction.
	// A value of 0 means they are perpendicular to each other.
	// So if value is negative, there won't be any diffuse or specular contributions
	// since the light source cannot hit those areas
	if( dot( lightSourceVector, normalVector ) > 0.0 ) {

		// Diffuse Contribution
		diffuseContribution = diffuse * dot( lightSourceVector, normalVector);

		// halfway vector, reduces the need for calculating a reflected vector,
		// which improvies performance slightly
		vec3 halfwayVector = normalize( viewVector + lightSourceVector);

		// Specular contribution
		specularContribution = specular * pow( dot(halfwayVector, normalVector), shininess);
	}

	// Calculate final color
	fragmentColor = vec4(lightIntensity * (ambientContribution + diffuseContribution + specularContribution), 1.0);
	
	//fragmentColor = vec4(0.4, 0.4, 0.4, 1.0);
		
	
	}
	else {

	if (hasTexture)
	{
		vec3 color = vec3(texture(Texture, TexCoords));
		vec3 ambientIntensity = vec3(0.3f, 0.3f, 0.3f);
		vec3 diffuseIntensity = vec3(0.4f, 0.4f, 0.4f);
		vec3 specularIntensity = vec3(0.5f, 0.5f, 0.5f);
		ambient = ambientIntensity * color;
		diffuse = diffuseIntensity * color;
		specular = specularIntensity * color;
		
	}
	
	// normal has been interpolated, so we need to normalize it
	vec3 normalVector = normalize(interpolatedNormal);

	// Calculate light source vector
	vec3 lightSourceVector = normalize( lightPosition.xyz - interpolatedPosition);

	// Calculate the view vector
	vec3 viewVector = normalize( -interpolatedPosition.xyz );

	// Ambient contribution
	vec3 ambientContribution = ambient;

	// Default black diffuse and specular contributions
	vec3 diffuseContribution = vec3(0.0);
	vec3 specularContribution = vec3(0.0);

	// Dot product of two normalized vectors gives us a value
	// indicating how close they are to each other,
	// A value of -1 means they are pointing in opposite directions
	// while a value of 1 means they are pointing in the same direction.
	// A value of 0 means they are perpendicular to each other.
	// So if value is negative, there won't be any diffuse or specular contributions
	// since the light source cannot hit those areas
	if( dot( lightSourceVector, normalVector ) > 0.0 ) {

		// Diffuse Contribution
		diffuseContribution = diffuse * dot( lightSourceVector, normalVector);

		// halfway vector, reduces the need for calculating a reflected vector,
		// which improvies performance slightly
		vec3 halfwayVector = normalize( viewVector + lightSourceVector);

		// Specular contribution
		specularContribution = specular * pow( dot(halfwayVector, normalVector), shininess);
	}

	// Calculate final color
	fragmentColor = vec4(lightIntensity * (ambientContribution + diffuseContribution + specularContribution), 1.0);
	
	}
    

}